﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using ZedGraph;
using System.Threading;

namespace lab4
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            comboBox1.SelectedIndex = 0;
            pixelFormat_label1.Text = "";
            pixelFormat_label2.Text = "";

            for (int i = 0; i < X_values.Length; i++)
                X_values[i] = i;
        }

        private double[] X_values = new double[256];
        string fileName;

        Bitmap bitmap = null;
        Bitmap bitmapRes = null;

        private void openButton_Click(object sender, EventArgs e)
        {
            OpenFileDialog dialog = new OpenFileDialog();
            dialog.Filter = "Image files (*.jpg, *.jpeg, *.jpe, *.jfif, *.png, *.bmp) | *.jpg; *.jpeg; *.jpe; *.jfif; *.png; *.bmp";

            if (dialog.ShowDialog() == DialogResult.OK)
            {
                if (bitmap != null)
                    bitmap.Dispose();

                if (bitmapRes != null)
                {
                    bitmapRes.Dispose();
                    bitmapRes = null;
                    pictureBox2.Image = null;
                    pixelFormat_label2.Text = "";
                    zedGraphControl2.GraphPane.CurveList.Clear();
                    zedGraphControl2.Invalidate();
                }

                Bitmap origBitmap = new Bitmap(dialog.FileName);
                fileName = System.IO.Path.GetFileNameWithoutExtension(dialog.FileName);
                pixelFormat_label1.Text = origBitmap.PixelFormat.ToString();

                if (origBitmap.PixelFormat != PixelFormat.Format24bppRgb && origBitmap.PixelFormat != PixelFormat.Format32bppArgb
                    && origBitmap.PixelFormat != PixelFormat.Format32bppRgb && origBitmap.PixelFormat != PixelFormat.Canonical)
                {
                    MessageBox.Show("Конвертирование!!!");
                    Bitmap cloneBitmap = new Bitmap(origBitmap.Width, origBitmap.Height, PixelFormat.Format24bppRgb);
                    using (Graphics gr = Graphics.FromImage(cloneBitmap))
                    {
                        gr.DrawImage(origBitmap, new Rectangle(0, 0, cloneBitmap.Width, cloneBitmap.Height));
                    }
                    origBitmap.Dispose();
                    origBitmap = cloneBitmap;
                }

                pictureBox1.Image = origBitmap;
                bitmap = origBitmap;
                drawHistogram(zedGraphControl1, bitmap, "Гистограмма изображения");
            }
        }

        private void solarizationButton_Click(object sender, EventArgs e)
        {
            if (bitmap == null)
                return;

            Bitmap inputBitmap;
            if (checkBox_result.Checked && bitmapRes != null)
                inputBitmap = bitmapRes;
            else inputBitmap = bitmap;

            double k = double.Parse(k_textBox.Text);
            bool isNormalization = checkBox1.Checked;

            bitmapRes = ImageProcessing.Solarisation(inputBitmap, isNormalization, k);

            showResult();
        }

        private void showResult()
        {
            if (bitmapRes == null)
                return;

            drawHistogram(zedGraphControl2, bitmapRes, "Гистограмма результата");
            pictureBox2.Image = bitmapRes;
            pixelFormat_label2.Text = bitmapRes.PixelFormat.ToString();
        }

        unsafe private void drawHistogram(ZedGraphControl zedGraphControl, Bitmap bitmapSrc, string title)
        {
            double[] frequency = new double[256];           
            int pos;

            int depth = Bitmap.GetPixelFormatSize(bitmapSrc.PixelFormat);

            int width = bitmapSrc.Width;
            int height = bitmapSrc.Height;

            UnsafeBitmap src_wrap = new UnsafeBitmap(bitmapSrc);
            Byte* pBaseSrc = src_wrap.LockBitmap();

            int i;
            byte r, g, b, a;

            if(depth == 24)
                for (int y = 0; y < height; y++)
                {
                    for (int x = 0; x < width; x++)
                    {
                        i = ((y * width) + x) * 3;
                        r = pBaseSrc[i];
                        g = pBaseSrc[i + 1];
                        b = pBaseSrc[i + 2];

                        pos = (int)(0.3 * r + 0.59 * g + 0.11 * b);
                        frequency[pos]++;
                    }
                }
            else if(depth == 32)
                for (int y = 0; y < height; y++)
                {
                    for (int x = 0; x < width; x++)
                    {
                        i = ((y * width) + x) * 4;
                        r = pBaseSrc[i];
                        g = pBaseSrc[i + 1];
                        b = pBaseSrc[i + 2];
                        a = pBaseSrc[i + 3];

                        if (a != 0)
                        {
                            pos = (int)(0.3 * r + 0.59 * g + 0.11 * b);
                            frequency[pos]++;
                        }                        
                    }
                }

            src_wrap.UnlockBitmap();

            // Получим панель для рисования
            GraphPane pane = zedGraphControl.GraphPane;
            pane.XAxis.Title.Text = "Значение яркости";
            pane.YAxis.Title.Text = "Частота";
            pane.Title.Text = title;

            // Очистим список кривых на тот случай, если до этого сигналы уже были нарисованы
            pane.CurveList.Clear();

            BarItem bar = pane.AddBar("", X_values, frequency, Color.DarkGreen);

            // !!! Расстояния между кластерами (группами столбиков) гистограммы = 0.0
            // У нас в кластере только один столбик.
            pane.BarSettings.MinClusterGap = 0.5f;
            //pane.BarSettings.ClusterScaleWidth = 0.005;
            bar.Bar.Border.IsVisible = false;
            bar.Bar.Fill = new Fill(Color.Blue);

            pane.XAxis.Scale.Min = X_values.First() - 10;
            pane.XAxis.Scale.Max = X_values.Last() + 10;
            pane.XAxis.Scale.AlignH = AlignH.Center;

            // Вызываем метод AxisChange (), чтобы обновить данные об осях. 
            zedGraphControl.AxisChange();

            // Обновляем график
            zedGraphControl.Invalidate();
        }

        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked)
                k_textBox.Enabled = false;
            else
                k_textBox.Enabled = true;
        }

        private void filterButton_Click(object sender, EventArgs e)
        {
            if (bitmap == null)
                return;

            Bitmap inputBitmap;
            if (checkBox_result.Checked && bitmapRes != null)
                inputBitmap = bitmapRes;
            else inputBitmap = bitmap;

            switch (comboBox1.SelectedIndex)
            {
                case 0:     // Тиснение (рельеф), Emboss
                    bitmapRes = ImageProcessing.Emboss(inputBitmap);
                    break;

                case 1:     // Бинарное псевдополутоновое (чёрно-белое)
                    bitmapRes = ImageProcessing.PseudoHalftone(inputBitmap, ImageProcessing.BinaryMaskType.D2, true);
                    break;

                case 2:     // Бинарное псевдополутоновое (цветное)
                    bitmapRes = ImageProcessing.PseudoHalftone(inputBitmap, ImageProcessing.BinaryMaskType.D4, false);
                    break;

                case 3:     // Низкочастотный фильтр
                    bitmapRes = ImageProcessing.LowpassFilter(inputBitmap, ImageProcessing.LowpassFilterеType.H1);
                    break;

                default:
                    break;
            }
            showResult();
        }

        private void saveButton_Click(object sender, EventArgs e)
        {
            //lockBitmapRes.GetBitMap().Save()
            if (bitmapRes == null)
                return;

            SaveFileDialog dialog = new SaveFileDialog();
            dialog.Filter = "Png Image (.png)|*.png|Gif Image (.gif)|*.gif|JPEG Image (.jpeg)|*.jpeg|Bitmap Image (.bmp)|*.bmp|Tiff Image (.tiff)|*.tiff";
            dialog.FileName = fileName;
            if (dialog.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    switch (dialog.FilterIndex)
                    {
                        case 1:      // png
                            bitmapRes.Save(dialog.FileName, ImageFormat.Png);
                            break;

                        case 2:     // gif
                            bitmapRes.Save(dialog.FileName, ImageFormat.Gif);
                            break;

                        case 3:     // jpeg
                            bitmapRes.Save(dialog.FileName, ImageFormat.Jpeg);
                            break;

                        case 4:    // bmp
                            bitmapRes.Save(dialog.FileName, ImageFormat.Bmp);
                            break;

                        case 5:    // tiff
                            bitmapRes.Save(dialog.FileName, ImageFormat.Tiff);
                            break;
                    }
                }
                catch (System.Exception ex)
                {
                    MessageBox.Show(ex.Message, "Ошибка сохранения", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                                
            }
        }

        private void pictureBox1_DoubleClick(object sender, EventArgs e)
        {
            frmPreview f = new frmPreview();
            f.PreviewImage = pictureBox1.Image;
            f.Show();
        }

        private void pictureBox2_DoubleClick(object sender, EventArgs e)
        {
            frmPreview f = new frmPreview();
            f.PreviewImage = pictureBox2.Image;
            f.Show();
        }

        
    }
}
